@@ -5,17 +5,21 @@ reference files (files that might have been plagairised from).
 
 from pathlib import Path
 import time
-import numpy as np
 import logging
-from .utils import (filter_code, highlight_overlap, get_copied_slices,
-                    get_document_fingerprints, find_fingerprint_overlap)
-import matplotlib.pyplot as plt
 import webbrowser
 import pkg_resources
-from jinja2 import Template
-from tqdm import tqdm
 import io
 import base64
+import warnings
+
+from tqdm import tqdm
+import numpy as np
+import matplotlib.pyplot as plt
+from jinja2 import Template
+
+from .utils import (filter_code, highlight_overlap, get_copied_slices,
+                    get_document_fingerprints, find_fingerprint_overlap)
+from . import defaults
 
 class CodeFingerprint:
     """Class for tokenizing, filtering, fingerprinting, and winnowing
@@ -33,9 +37,13 @@ class CodeFingerprint:
     boilerplate : array_like, optional
         List of fingerprints to use as boilerplate. Any fingerprints
         present in this list will be discarded from the hash list.
-    filter : bool
+    filter : bool, default=True
         If set to to False, code will not be tokenized & filtered.
-        Default: True
+    fp : TextIO, default=None
+        I/O stream for data to create a fingerprint for. If provided,
+        the "file" argument will not be used to load a file from disk
+        but will still be used for language detection and displayed on
+        the report.
 
     Attributes
     ----------
@@ -62,9 +70,12 @@ class CodeFingerprint:
         rather than guessing from the file extension.
     """
     def __init__(self, file, k, win_size, boilerplate=[], filter=True,
-                 language=None):
-        with open(file) as code_fp:
-            code = code_fp.read()
+                 language=None, fp=None):
+        if fp is not None:
+            code = fp.read()
+        else:
+            with open(file) as code_fp:
+                code = code_fp.read()
         if filter:
             filtered_code, offsets = filter_code(code, file, language)
         else:
@@ -119,11 +130,11 @@ def compare_files(file1_data, file2_data):
     token_overlap2 = np.sum(slices2[1] - slices2[0])
 
     if len(file1_data.filtered_code) > 0:
-        similarity1 = token_overlap1 / len(file1_data.filtered_code)
+        similarity1 = len(idx1) / len(file1_data.hashes)
     else:
         similarity1 = 0
     if len(file2_data.filtered_code) > 0:
-        similarity2 = token_overlap2 / len(file2_data.filtered_code)
+        similarity2 = len(idx2) / len(file2_data.hashes)
     else:
         similarity2 = 0
 
@@ -151,6 +162,9 @@ class CopyDetector:
         below. If provided, parameters set in the configuration
         dictionary will overwrite default parameters and other
         parameters passed to the initialization function.
+
+        Note that this parameter is deprecated and will be removed in a
+        future version.
     test_dirs : list
         (test_directories) A list of directories to recursively search
         for files to check for plagiarism.
@@ -206,10 +220,27 @@ class CopyDetector:
     """
     def __init__(self, config=None, test_dirs=[], ref_dirs=[],
                  boilerplate_dirs=[], extensions=["*"],
-                 noise_t=25, guarantee_t=30, display_t=0.33,
+                 noise_t=defaults.NOISE_THRESHOLD,
+                 guarantee_t=defaults.GUARANTEE_THRESHOLD,
+                 display_t=defaults.DISPLAY_THRESHOLD,
                  same_name_only=False, ignore_leaf=False, autoopen=True,
                  disable_filtering=False, force_language=None,
                  truncate=False, out_file="./report.html", silent=False):
+        if config is not None:
+            # temporary workaround to ensure existing code continues
+            # to work
+            warnings.warn(
+                "use CopyDetector.from_config to initialize CopyDetector from "
+                "a config file. The config parameter is deprecated and will be"
+                " removed in a future version.",
+                DeprecationWarning, stacklevel=2
+            )
+            args = locals()
+            del args["self"]
+            del args["config"]
+            self.__init__(**{**args, **self._read_config(config)})
+            return
+
         self.silent = silent
         self.test_dirs = test_dirs
         if len(ref_dirs) == 0:
@@ -229,9 +260,6 @@ class CopyDetector:
         self.truncate = truncate
         self.out_file = out_file
 
-        if config is not None:
-            self._load_config(config)
-
         self._check_arguments()
 
         out_path = Path(self.out_file)
@@ -247,6 +275,52 @@ class CopyDetector:
         self.boilerplate_files = self._get_file_list(self.boilerplate_dirs,
                                                      self.extensions)
 
+        # before run() is called, similarity data should be empty
+        self.similarity_matrix = np.array([])
+        self.token_overlap_matrix = np.array([])
+        self.slice_matrix = np.array([])
+        self.file_data = {}
+
+    @staticmethod
+    def _read_config(config: dict) -> dict:
+        """Helper function for translating json configuration parameters
+        to the arguments taken by __init__
+        """
+        config_param_mapping = {
+            "noise_threshold": "noise_t",
+            "guarantee_threshold": "guarantee_t",
+            "display_threshold": "display_t",
+            "test_directories": "test_dirs",
+            "reference_directories": "ref_dirs",
+            "boilerplate_directories": "boilerplate_dirs"
+        }
+        for conf_key, param_key in config_param_mapping.items():
+            if conf_key in config:
+                config[param_key] = config[conf_key]
+                del config[conf_key]
+        if "disable_autoopen" in config:
+            config["autoopen"] = not config["disable_autoopen"]
+            del config["disable_autoopen"]
+
+        return config
+
+    @classmethod
+    def from_config(cls, config):
+        """Initializes a CopyDetection object using the provided
+        configuration dictionary.
+
+        Parameters
+        ----------
+        config : dict
+            Configuration dictionary using CLI parameter names.
+
+        Returns
+        -------
+        CopyDetector
+            Detection object initialized with config
+        """
+        return cls(**cls._read_config(config))
+
     def _load_config(self, config):
         """Sets member variables according to a configuration
         dictionary.
@@ -318,11 +392,11 @@ class CopyDetector:
                              "equal to noise threshold")
         if self.display_t > 1 or self.display_t < 0:
             raise ValueError("Display threshold must be between 0 and 1")
-        if Path(self.out_file).parent.exists() == False:
+        if not Path(self.out_file).parent.exists():
             raise ValueError("Invalid output file path "
                 "(directory does not exist)")
 
-    def _get_file_list(self, dirs, exts, unique=True):
+    def _get_file_list(self, dirs, exts):
         """Recursively collects list of files from provided
         directories. Used to search test_dirs, ref_dirs, and
         boilerplate_dirs
@@ -340,23 +414,30 @@ class CopyDetector:
                     logging.warning("No files found in " + dir)
                 file_list.extend(files)
 
-        return set(file_list)
+        # convert to a set to remove duplicates, then back to a list
+        return list(set(file_list))
 
     def add_file(self, filename, type="testref"):
         """Adds a file to the list of test files, reference files, or
-        boilerplate files. The "type" parameter should be one of
-        ["testref", "test", "ref", "boilerplate"]. "testref" will add
-        the file as both a test and reference file.
+        boilerplate files.
+
+        Parameters
+        ----------
+        filename : str
+            Name of file to add.
+        type : {"testref", "test", "ref", "boilerplate"}
+            Type of file to add. "testref" will add the file as both a
+            test and reference file.
         """
         if type == "testref":
-            self.test_files.add(filename)
-            self.ref_files.add(filename)
+            self.test_files.append(filename)
+            self.ref_files.append(filename)
         elif type == "test":
-            self.test_files.add(filename)
+            self.test_files.append(filename)
         elif type == "ref":
-            self.ref_files.add(filename)
+            self.ref_files.append(filename)
         elif type == "boilerplate":
-            self.boilerplate_files.add(filename)
+            self.boilerplate_files.append(filename)
 
     def _get_boilerplate_hashes(self):
         """Generates a list of hashes of the boilerplate text. Returns
@@ -366,35 +447,34 @@ class CopyDetector:
         boilerplate_hashes = []
         for file in self.boilerplate_files:
             try:
-                with open(file) as boilerplate_fp:
-                    boilerplate = boilerplate_fp.read()
+                fingerprint=CodeFingerprint(file, self.noise_t, 1,
+                                            filter=not self.disable_filtering,
+                                            language=self.force_language)
+                boilerplate_hashes.extend(fingerprint.hashes)
             except UnicodeDecodeError:
                 logging.warning(f"Skipping {file}: file not ASCII text")
                 continue
-            fingerprint = CodeFingerprint(file, self.noise_t, 1,
-                                          filter=not self.disable_filtering,
-                                          language=self.force_language)
-            boilerplate_hashes.extend(fingerprint.hashes)
 
         return np.unique(np.array(boilerplate_hashes))
 
     def _preprocess_code(self, file_list):
+        """Generates a CodeFingerprint object for each file in the
+        provided file list. This is where the winnowing algorithm is
+        actually used.
+        """
         boilerplate_hashes = self._get_boilerplate_hashes()
-
-        file_data = {}
         for code_f in tqdm(file_list, bar_format= '   {l_bar}{bar}{r_bar}',
-                disable=self.silent):
-            try:
-                file_data[code_f] = CodeFingerprint(
-                    code_f, self.noise_t, self.window_size,
-                    boilerplate_hashes, not self.disable_filtering,
-                    self.force_language)
+                           disable=self.silent):
+            if code_f not in self.file_data:
+                try:
+                    self.file_data[code_f] = CodeFingerprint(
+                        code_f, self.noise_t, self.window_size,
+                        boilerplate_hashes, not self.disable_filtering,
+                        self.force_language)
 
-            except UnicodeDecodeError:
-                logging.warning(f"Skipping {code_f}: file not ASCII text")
-                continue
-
-        return file_data
+                except UnicodeDecodeError:
+                    logging.warning(f"Skipping {code_f}: file not ASCII text")
+                    continue
 
     def _comparison_loop(self):
         """The core code used to determine code overlap. The overlap
@@ -406,97 +486,101 @@ class CopyDetector:
         start_time = time.time()
         if not self.silent:
             print("  0.00: Generating file fingerprints")
-        test_f_list = sorted(list(self.test_files))
-        self.all_files = (test_f_list
-            + sorted([f for f in self.ref_files if f not in self.test_files]))
-        self.file_data = self._preprocess_code(self.all_files)
+        self._preprocess_code(self.test_files + self.ref_files)
 
         self.similarity_matrix = np.full((
-            len(self.all_files), len(self.all_files)), -1, dtype=np.float64)
+            len(self.test_files),len(self.ref_files),2), -1, dtype=np.float64)
         self.token_overlap_matrix = np.full((
-            len(self.all_files), len(self.all_files)), -1)
-        self.slice_matrix = [[np.array([]) for _ in range(len(self.all_files))]
-                             for _ in range(len(self.all_files))]
+            len(self.test_files), len(self.ref_files)), -1)
+        self.slice_matrix = [[np.array([]) for _ in range(len(self.ref_files))]
+                             for _ in range(len(self.test_files))]
 
         if not self.silent:
             print(f"{time.time()-start_time:6.2f}: Beginning code comparison")
 
-        for i, test_f in enumerate(tqdm(test_f_list,
+        # this is used to track which files have been compared to avoid
+        # unnecessary duplication when there is overlap between the
+        # test and reference files
+        comparisons = {}
+
+        for i, test_f in enumerate(tqdm(self.test_files,
                 bar_format= '   {l_bar}{bar}{r_bar}', disable=self.silent)):
-            for j, ref_f in enumerate(self.all_files):
-                if test_f not in self.file_data or ref_f not in self.file_data:
-                    continue
-                elif test_f == ref_f:
-                    continue
-                elif self.similarity_matrix[i,j] != -1:
-                    continue
-                elif (self.all_files[i] not in self.test_files or
-                      self.all_files[j] not in self.ref_files):
+            for j, ref_f in enumerate(self.ref_files):
+                if (test_f not in self.file_data
+                        or ref_f not in self.file_data
+                        or test_f == ref_f
+                        or (self.same_name_only
+                            and (Path(test_f).name != Path(ref_f).name))
+                        or (self.ignore_leaf
+                            and (Path(test_f).parent == Path(ref_f).parent))):
                     continue
 
-                if self.same_name_only:
-                    if Path(test_f).name != Path(ref_f).name:
-                        continue
-                if self.ignore_leaf:
-                    if Path(test_f).parent == Path(ref_f).parent:
-                        continue
+                if (ref_f, test_f) in comparisons:
+                    ref_idx, test_idx = comparisons[(ref_f, test_f)]
+                    overlap = self.token_overlap_matrix[ref_idx, test_idx]
+                    sim2, sim1 = self.similarity_matrix[ref_idx, test_idx]
+                    slices2, slices1 = self.slice_matrix[ref_idx][test_idx]
+                else:
+                    overlap, (sim1, sim2), (slices1, slices2) = compare_files(
+                        self.file_data[test_f], self.file_data[ref_f])
+                    comparisons[(test_f, ref_f)] = (i, j)
 
-                overlap, (sim1,sim2), (slices1,slices2) = compare_files(
-                    self.file_data[test_f], self.file_data[ref_f])
-
-                self.similarity_matrix[i,j] = sim1
+                self.similarity_matrix[i,j] = np.array([sim1, sim2])
                 self.slice_matrix[i][j] = [slices1, slices2]
-                self.similarity_matrix[j,i] = sim2
-                self.slice_matrix[j][i] = [slices2,slices1]
-
                 self.token_overlap_matrix[i,j] = overlap
-                self.token_overlap_matrix[j,i] = overlap
 
         if not self.silent:
             print(f"{time.time()-start_time:6.2f}: Code comparison completed")
 
     def run(self):
-        """User-facing code overlap computing function. Checks for a
-        session that can be resumed from then calls _comparison_loop to
-        generate results.
+        """Runs the copy detection loop for detecting overlap between
+        test and reference files. If no files are in the provided
+        directories, the similarity matrix will remain empty and any
+        attempts to generate a report will fail.
         """
-        if len(self.test_files) == 0 or len(self.ref_files) == 0:
-            err_folder = "test"
-            if len(self.test_files) > len(self.ref_files):
-                err_folder = "reference"
-
+        if len(self.test_files) == 0:
             logging.error("Copy detector failed: No files found in "
-                          f"{err_folder} directories")
-            self.similarity_matrix = np.array([])
-            self.token_overlap_matrix = np.array([])
-            self.slice_matrix = np.array([])
-            return
-
-        self._comparison_loop()
+                          "test directories")
+        elif len(self.ref_files) == 0:
+            logging.error("Copy detector failed: No files found in "
+                          "reference directories")
+        else:
+            self._comparison_loop()
 
     def get_copied_code_list(self):
         """Get a list of copied code to display on the output report.
         Returns a list of tuples containing the similarity score, the
         test file name, the compare file name, the highlighted test
-        code, and the highlighted compare code,
+        code, and the highlighted compare code.
+
+        Returns
+        -------
+        list
+            list of similarity data between each file pair which
+            achieves a similarity score above the display threshold,
+            ordered by percentage of copying in the test file. Each
+            element of the list contains [test similarity, reference
+            similarity, path to test file, path to reference file,
+            highlighted test code, highlighted reference code, numer of
+            overlapping tokens]
         """
         if len(self.similarity_matrix) == 0:
             logging.error("Cannot generate code list: no files compared")
             return []
-        x,y = np.where(self.similarity_matrix > self.display_t)
+        x,y = np.where(self.similarity_matrix[:,:,0] > self.display_t)
 
         code_list = []
-        selected_pairs = set([])
+        file_pairs = set()
         for idx in range(len(x)):
-            test_f = self.all_files[x[idx]]
-            ref_f = self.all_files[y[idx]]
-            if test_f + ref_f in selected_pairs:
+            test_f = self.test_files[x[idx]]
+            ref_f = self.ref_files[y[idx]]
+            if (ref_f, test_f) in file_pairs:
+                # if comparison is already in report, don't add it again
                 continue
+            file_pairs.add((test_f, ref_f))
 
-            selected_pairs.add(test_f + ref_f)
-            selected_pairs.add(ref_f + test_f)
-            test_sim = self.similarity_matrix[x[idx], y[idx]]
-            ref_sim = self.similarity_matrix[y[idx], x[idx]]
+            test_sim = self.similarity_matrix[x[idx], y[idx], 0]
+            ref_sim = self.similarity_matrix[x[idx], y[idx], 1]
             slices_test = self.slice_matrix[x[idx]][y[idx]][0]
             slices_ref = self.slice_matrix[x[idx]][y[idx]][1]
 
@@ -524,6 +608,13 @@ class CopyDetector:
         """Generates an html report listing all files with similarity
         above the display_threshold, with the copied code segments
         highlighted.
+
+        Parameters
+        ----------
+        output_mode : {"save", "return"}
+            If "save", the output will be saved to the file specified
+            by self.out_file. If "return", the output HTML will be
+            directly returned by this function.
         """
         if len(self.similarity_matrix) == 0:
             logging.error("Cannot generate report: no files compared")
@@ -532,7 +623,7 @@ class CopyDetector:
         code_list = self.get_copied_code_list()
         data_dir = pkg_resources.resource_filename('copydetect', 'data/')
 
-        plot_mtx = np.copy(self.similarity_matrix)
+        plot_mtx = np.copy(self.similarity_matrix[:,:,0])
         plot_mtx[plot_mtx == -1] = np.nan
         plt.imshow(plot_mtx)
         plt.colorbar()
@@ -543,7 +634,7 @@ class CopyDetector:
         sim_mtx_base64 = base64.b64encode(sim_mtx_buffer.read()).decode()
         plt.close()
 
-        scores = self.similarity_matrix[self.similarity_matrix != -1]
+        scores=self.similarity_matrix[:,:,0][self.similarity_matrix[:,:,0]!=-1]
         plt.hist(scores, bins=20)
         plt.tight_layout()
         sim_hist_buffer = io.BytesIO()
@@ -556,7 +647,7 @@ class CopyDetector:
         with open(data_dir + "report.html") as template_fp:
             template = Template(template_fp.read())
 
-        flagged = self.similarity_matrix > self.display_t
+        flagged = self.similarity_matrix[:,:,0] > self.display_t
         flagged_file_count = np.sum(np.any(flagged, axis=1))
 
         output = template.render(test_count=len(self.test_files),
